package xyz.marsavic.gfxlab.graphics3d.raytracers;

import xyz.marsavic.gfxlab.Color;
import xyz.marsavic.gfxlab.Vec3;
import xyz.marsavic.gfxlab.graphics3d.*;

public class RayTracerSimple extends RayTracer {

	private static final int spectrumSamples = 20;
	private static final double minWavelength = 380;
	private static final double maxWavelength = 780;
	private static final double EPSILON = 1e-9;

	public RayTracerSimple(Scene scene, Camera camera) {
		super(scene, camera);
	}
	
	@Override
	protected Color sample(Ray ray) {
		return sample(ray, 64);
	}

	protected Color sample(Ray ray, int depthRemaining) {
		double x = 0.0, y = 0.0, z = 0.0;
		for (int i = 0; i < spectrumSamples; i++) {
			double wavelength = minWavelength + (((double) i)/spectrumSamples)*(maxWavelength - minWavelength);
			double intensity = sample(ray, depthRemaining, wavelength);

			x += intensity * Xyz.x[(int) wavelength];
			y += intensity * Xyz.y[(int) wavelength];
			z += intensity * Xyz.z[(int) wavelength];
		}

		Color result = Color.xyz(x,y,z);

		return result;
	}

	protected double sample(Ray ray, int depthRemaining, double wavelength) {
		if (depthRemaining == 0) {
			return 0.0;
		}
		
		Hit hit = scene.solid().firstHit(ray, EPSILON);
		if (hit == null) {
			return scene.backgroundSpectrum.at(wavelength);
		}
		
		Vec3 p = ray.at(hit.t());                 // The hit point
		Vec3 n_ = hit.n_();                       // Normalized normal to the body surface at the hit point
		Vec3 i = ray.d().inverse();               // Incoming direction
		double lI = i.length();
		Vec3 r = GeometryUtils.reflectedN(n_, i); // Reflected ray (i reflected over n)
		Vec3 r_ = r.div(lI);                      // Reflected ray (i reflected over n)
		
		double lightDiffuse  = 0.0;        // The sum of diffuse contributions from all the lights
		double lightSpecular = 0.0;        // The sum of specular contributions from all the lights
		
		Material material = hit.material();
		
		for (Light light : scene.lights()) {
			Vec3 l = light.p().sub(p);            // Vector from p to the light;
			
			Ray rayToLight = Ray.pd(p, l);
			if (scene.solid().hitBetween(rayToLight, EPSILON, 1)) continue;
			
			double lLSqr = l.lengthSquared();     // Distance from p to the light squared
			double lL = Math.sqrt(lLSqr);         // Distance from p to the light
			double cosLN = n_.dot(l) / lL;        // Cosine of the angle between l and n_
			
			if (cosLN > 0) {                      // If the light is above the surface
				double irradiance = light.s().at(wavelength) * cosLN / lLSqr;
				// The irradiance represents how much light is received by a unit area of the surface. It is
				// proportional to the cosine of the incoming angle and inversely proportional to the distance squared
				// (inverse-square law).
				lightDiffuse = lightDiffuse+irradiance;
				
				double cosLR = l.dot(r_);
				if (cosLR > 0) {                  // If the angle between l and r is acute
					cosLR /= lL;
					lightSpecular = lightSpecular + irradiance * Math.pow(cosLR, material.shininess());
				}
			}
		}
		
		double result = 0.0;
		result += material.diffuse().at(wavelength)  * lightDiffuse;
		result += material.specular().at(wavelength) * lightSpecular;

		double reflective = material.reflective().at(wavelength);
		if (reflective != 0.0) {
			// When material has reflective properties we recursively find the color visible along the ray (p, r).
			double lightReflected = sample(Ray.pd(p, r), depthRemaining - 1, wavelength);
			result += reflective * lightReflected;
		}

		double refractive = material.refractive().at(wavelength);
		if (refractive != 0.0) {
			Vec3 b; // refracted light vector
			double rInd = 1/material.refractiveIndex().at(wavelength);

			double iCosN = i.dot(n_);
			if (iCosN < 0) {
				rInd = 1/rInd;
			}
			Vec3 iProjection = n_.mul(iCosN);
			Vec3 iRejection = i.sub(iProjection);
			double iSinSqr = iRejection.lengthSquared()/i.lengthSquared();
			double bSinSqr = iSinSqr*rInd*rInd;
			if (bSinSqr > 1) {
				b = r;
			} else {
				Vec3 bRejection = iRejection.inverse();
				Vec3 bProjection = n_.mul(Math.sqrt(bRejection.lengthSquared()*(1-bSinSqr)/bSinSqr));
				if (iCosN > 0) {
					bProjection = bProjection.inverse();
				}
				b = bRejection.add(bProjection);
			}
			double lightRefracted = sample(Ray.pd(p, b), depthRemaining - 1, wavelength);
			result += refractive * lightRefracted;
		}
		
		return result;
	}
}
